/*
 This file is part of the digitalSTROM ds485-client library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM ds485-client library is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 digitalSTROM ds485-client library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with digitalSTROM ds485-client library. If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef DS485CLIENT_H_
#define DS485CLIENT_H_

#include <stdbool.h>

#include <digitalSTROM/ds485-core.h>
#include "ds485-client-interface.h"

#if HAVE_VISIBILITY
#define PUBLIC_API(sym) typeof(sym) sym __attribute__((__visibility__("default")));
#else
#define PUBLIC_API(sym)
#endif

typedef enum
{
    STATE_IDLE,
    STATE_WAITING,
    STATE_RECEIVED,
} ds485_client_response_state_t;

typedef struct {
    ds485_client_response_state_t state;
    pthread_mutex_t mutex;
    pthread_cond_t condition;
} waitqueue_t;

typedef struct
{
    waitqueue_t wq;
    ds485_container_t response;
} ds485_client_syncsend_t;

typedef struct
{
    waitqueue_t wq;
    ds485n_packet_t response;
} ds485_client_syncsend_long_t;

typedef struct
{
    waitqueue_t wq;
    uint8_t length;
    dsuid_t *devices;
    enum {
        query_dsuid,
        query_devices,
        query_state,
        restart_bus
    } type;
} ds485_client_sync_bus_query_t;

typedef struct {
    waitqueue_t wq;
} ds485_client_open_query_t;

typedef struct _ds485_client_s
{
    bus_state_t busState;
    int flags;

    char pretty_name[30];

    bool shutdown;
    int wait_pong;
    pthread_t connThread;
    ds485_client_syncsend_t syncSendShort;
    ds485_client_syncsend_long_t syncSendLong;
    ds485_client_sync_bus_query_t busQuery;
    ds485_client_open_query_t openQuery;

    int last_msg_opcode;
    int last_msg_len;

    /* socket specific */
    char *dProtocol;
    char *dName;
    char *dPort;
    int sockfd;

    /* dsm-api callbacks */
    struct ds485c_callbacks callbacks;
} ds485_client_t;

extern void waitqueue_create(waitqueue_t *wq);
extern void waitqueue_destroy(waitqueue_t *wq);
extern int waitqueue_enqueue(waitqueue_t *wq);
extern int waitqueue_wait(waitqueue_t *wq, int timeout_sec);
extern int waitqueue_dequeue(waitqueue_t *wq);

/**
 * waitqueue_waiting
 * @return true if thread is waiting for signal
 * must pair with waitqueue_unlock
 */
extern bool waitqueue_waiting(waitqueue_t *wq);
extern int waitqueue_signal(waitqueue_t *wq);
extern int waitqueue_unlock(waitqueue_t *wq);

extern int ds485_socket_client_open(ds485_client_t *cl);
extern int ds485_socket_client_send_command(ds485_client_t *cl,
                                            const ds485_container_t *request);
extern int ds485_socket_client_send_long_command(ds485_client_t *cl,
                                                 const ds485n_packet_t *request);
extern int ds485_socket_client_get_bus_devices(ds485_client_t *cl, dsuid_t *device_list, int device_list_size);
extern int ds485_socket_client_get_bus_state(ds485_client_t *cl, bus_state_t *state);
extern int ds485_socket_client_get_dsuid(ds485_client_t *cl);
extern int ds485_socket_client_restart_bus(ds485_client_t *cl);

extern int dlog(int level, struct _ds485_client_s *cl, const char *fmt, ...);

extern int do_container_received(ds485_client_t *cl, ds485_container_t *cnt);
extern int do_long_container_received(ds485_client_t *cl, ds485n_packet_t *pkt);
extern int do_bus_state_event_received(ds485_client_t *cl, bus_state_t state);
extern int do_bus_state_query_received(ds485_client_t *cl, bus_state_t state);
extern int do_device_list_received(ds485_client_t *cl, uint8_t *octet, int len);
extern int do_bus_change_received(ds485_client_t *cl, dsuid_t dsuid, int flags);
extern int do_dsid_query_received(ds485_client_t *cl, uint8_t *octet);
extern int do_restart_bus_received(ds485_client_t *cl);
extern int do_client_connect_success(ds485_client_t *cl);

extern void update_pretty_name(ds485_client_t *cl);

#endif /* DS485CLIENT_H_ */
